""" Distributor init file

Distributors: you can add custom code here to support particular distributions
of scikit-learn.

For example, this is a good place to put any checks for hardware requirements.

The scikit-learn standard source distribution will not put code in this file,
so you can safely replace this file with your own version.
"""
